-------------------------------------------------------------------------------
-- SmartFrameMover
-- Created by Aeldra (EU-Proudmoore)
--
-- Makes special frames movable.
-------------------------------------------------------------------------------

local AddonVersion       = "v6.0b";
local AddonTitle         = "SmartFrameMover";
local AddonTitleVersion  = AddonTitle.." "..AddonVersion;
local EventOnUpdate      = "SFMONUPDATE";

local O;
local _;
local fMain, fMover, fFocus;
local isInit = false;
local isCombat = false;
local AnchorPoint = "TOPLEFT";
local curFrame = "Boss1TargetFrame";

local HookObjectiveTracker_Update = ObjectiveTracker_Update;

local FrameList = {
  {"Boss1TargetFrame", "Boss Target"},
  {"VehicleSeatIndicator", "Vehicle Seat"},
  {"ObjectiveTrackerFrame", "Quest Tracker"},
  {"QuestTimerFrame", "Quest Timer"},  
  {"PlayerPowerBarAlt", "Special Power Bar"},
  {"ExtraActionButton1", "Extra Action Button 1"},
  {"CompactRaidFrameManager", "Compact Raid"},
  {"CompactRaidFrameManagerToggleButton", "Compact Raid"},
  {"GameTooltip", "Tooltip"},
  {"-", "-"},
	{"AchievementAlertFrame1", "Achievement Alert"},
	{"AlertFrame", "Alerts"},
  {"BuffFrame", "Buffs"},
  {"CastingBarFrame", "Casting Bar"},
  {"DurabilityFrame", "Durability Doll"},
  {"GroupLootFrame1", "Loot Roll"},
  {"MinimapCluster", "Minimap"},
  {"TicketStatusFrame", "Ticket Status"},
  {"StreamingIcon", "Streaming Icon"},
  {"WorldStateAlwaysUpFrame", "World State"},
  {"-", "-"},
  {"PlayerFrame", "Player"},
  {"TargetFrame", "Target"},
  {"TargetFrameToT", "Target of Target"},
  {"FocusFrame", "Focus"},
  {"FocusFrameToT", "Target of Focus"},
  {"PartyMemberFrame1", "Party Members"},
  {"ArenaEnemyFrame1", "Arena Enemies"},
}


-----------------------------------------------------------------------------------------------------------------
-- Local helper functions ---------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------

-- Rounds a number to the given number of decimal places. 
local r_mult;
local function Round(num, idp)
  r_mult = 10^(idp or 0);
  return math.floor(num * r_mult + 0.5) / r_mult;
end

-- Returns a chat color code string
local function BCC(r, g, b)
  return string.format("|cff%02x%02x%02x", (r*255), (g*255), (b*255));
end

local BL  = BCC(0.1, 0.1, 1.0);
local BLD = BCC(0.0, 0.0, 0.7);
local BLL = BCC(0.5, 0.8, 1.0);
local GR  = BCC(0.1, 1.0, 0.1);
local GRD = BCC(0.0, 0.7, 0.0);
local GRL = BCC(0.25, 0.75, 0.25);
local RD  = BCC(1.0, 0.1, 0.1);
local RDD = BCC(0.7, 0.0, 0.0);
local RDL = BCC(1.0, 0.3, 0.3);
local YL  = BCC(1.0, 1.0, 0.0);
local YLD = BCC(0.7, 0.7, 0.0);
local YLL = BCC(1.0, 1.0, 0.5);
local OR  = BCC(1.0, 0.5, 0.25);
local ORD = BCC(0.7, 0.5, 0.0);
local ORL = BCC(1.0, 0.6, 0.3);
local WH  = BCC(1.0, 1.0, 1.0);
local CY  = BCC(0.5, 1.0, 1.0);
local GY  = BCC(0.5, 0.5, 0.5);
local GYD = BCC(0.35, 0.35, 0.35);
local GYL = BCC(0.65, 0.65, 0.65);
local GFN = BCC(1.0, 0.82, 0.0);
local GFH = BCC(1.0, 1.0, 1.0);

-- Returns "" instead of nil
local function ChkS(s)
  if (s == nil) then
    s = "";
  end
  return s;
end

-- Converts a string to number
local function StrToNum(s, nDef, nMin, nMax)
  local n = nDef;
  if (s ~= nil) then
    n = tonumber(s);
    if (n == nil) then
      n = nDef;
    elseif (nMin ~= nil and n < nMin) then
      n = nMin;
    elseif (nMax ~= nil and n > nMax) then
      n = nMax;
    end
  end
  return n;
end

local function BToStr(b)
  if (b) then
    return GR.."On|r";
  else
    return RD.."Off|r";
  end
end

-- Splits a string
local function Split(msg, char)
  local arr = { };
  while (string.find(msg, char)) do
    local iStart, iEnd = string.find(msg, char);
    tinsert(arr, strsub(msg, 1, iStart - 1));
    msg = strsub(msg, iEnd + 1, strlen(msg));
  end
  if (strlen(msg) > 0) then
    tinsert(arr, msg);
  end
  return arr;
end

-- Will dump the value of msg to the default chat window
local function AddMsg(msg)
  if (DEFAULT_CHAT_FRAME) then
    DEFAULT_CHAT_FRAME:AddMessage(YLL..msg.."|r");
  end
end

local function AddMsgErr(msg)
  if (DEFAULT_CHAT_FRAME) then
    DEFAULT_CHAT_FRAME:AddMessage(RDL..AddonTitle..": "..msg.."|r");
  end
end

local function AddMsgWarn(msg)
  if (DEFAULT_CHAT_FRAME) then
    DEFAULT_CHAT_FRAME:AddMessage(CY..msg.."|r");
  end
end

local function AddMsgD(msg, r, g, b)
  if (r == nil) then r = 0.5; end
  if (g == nil) then g = 0.8; end
  if (b == nil) then b = 1; end
  if (DEFAULT_CHAT_FRAME and O and O.Debug) then
    DEFAULT_CHAT_FRAME:AddMessage(msg, r, g, b);
  end
end

local function BToggle(b, msg)
  if (not b or b == nil) then
    b = true;
    AddMsg(AddonTitle..": "..msg..GR.."On");
  else
    b = false
    AddMsg(AddonTitle..": "..msg..RD.."Off");
  end
  return b;
end

local function BState(b, msg)
  if (b) then
    AddMsg(AddonTitle..": "..msg..GR.."On");
  else
    AddMsg(AddonTitle..": "..msg..RD.."Off");
  end
end


local function ReplaceFunctions(f)
  if (f and not f.SFMIsReplaced) then
	  
	  --print(f:GetName().." replaced");
	  
	  if (f.SetPoint ~= nil and f.SFMSetPoint == nil) then
	    local point, relativeTo, relativePoint, xOfs, yOfs = f:GetPoint();
	    f.SFMOrigPoints = {point, relativeTo, relativePoint, xOfs, yOfs, f:GetScale(), f:GetAlpha()};
		  f.SFMSetPoint = f.SetPoint;
		  --AddMsgD(f:GetName().." SetPoint replaced");
		  f.SetPoint = function()
		    if (O == nil) then return end
		    
		    local v = O.Frames[f:GetName()];
		    if (v == nil) then return end
		    
        local s = 1;
        if (f ~= fMover) then s = v.S end
        f:SetScale(s);
        f:SFMSetPoint(AnchorPoint, UIParent, v.P, v.X/s, v.Y/s);
		  end;
	  end
  	
  	--[[
	  if (f.ClearAllPoints ~= nil and f.SFMClearAllPoints == nil) then
		  f.SFMClearAllPoints = f.ClearAllPoints;		
		  --AddMsgD(f:GetName().." ClearAllPoints replaced");
		  f.ClearAllPoints = function() end;
	  end
	  ]]--
  	
	  if (f.SetParent ~= nil and f.SFMSetParent == nil) then
	    f.OldParent = f:GetParent();
		  f.SFMSetParent = f.SetParent;		
		  f:SFMSetParent(UIParent);
		  --AddMsgD(f:GetName().." SetSetParent replaced");
		  f.SetParent = function()
		    f:SFMSetParent(UIParent);
		  end;
	  end
	  
	  if (f.SetOwner ~= nil and f.SFMSetOwner == nil) then
		  f.SFMSetOwner = f.SetOwner;		
	    f.SetOwner = function(tooltip, parent, ...)
  	    f:SFMSetOwner(parent, "ANCHOR_NONE");
  	    f:ClearAllPoints();
  	    f:SetPoint();
	    end;
	  end
	  
	  f.SFMIsReplaced = true;
	  AddMsgD(f:GetName().." functions replaced");
	end
end

local function RestoreFunctions(f)
  if (f and f.SFMIsReplaced and f ~= fMover) then
	  if (f.SFMSetPoint ~= nil) then
		  f.SetPoint = f.SFMSetPoint;
		  f.SFMSetPoint = nil;
	  end
	  if (f.SFMClearAllPoints ~= nil) then
		  f.ClearAllPoints = f.SFMClearAllPoints;
		  f.SFMClearAllPoints = nil;
	  end
	  if (f.SFMSetParent ~= nil) then
		  f.SetParent = f.SFMSetParent;
		  f:SetParent(f.OldParent);
		  f.SFMSetParent = nil;
	  end
	  if (f.SFMSetOwner ~= nil) then
		  f.SetOwner = f.SFMSetOwner;
		  f.SFMSetOwner = nil;
	  end	  
	  local v = f.SFMOrigPoints;
	  if (v) then
	    f:ClearAllPoints();
	    f:SetPoint(v[1], v[2], v[3], v[4], v[5]);
	    f:SetScale(v[6]);
      if (v[7]) then f:SetAlpha(v[7]) end;
	  end
	  f.SFMIsReplaced = false;
	  AddMsgD(f:GetName().." functions restored");
	end
end


local function UpdateFramePos(f, v)
  --if (InCombatLockdown()) then return; end
  if (f == nil or v == nil) then return; end
  
  if (v.MI) then
    ReplaceFunctions(f);
    --f:SetMovable(true);
    
    local s = 1;
    if (f ~= fMover) then s = v.S end
    f:ClearAllPoints();
    f:SetScale(s);
    f:SFMSetPoint(AnchorPoint, UIParent, v.P, v.X/s, v.Y/s);
    
    --f:ClearAllPoints();
    --f:SetPoint();
    
    --AddMsgD(f:GetName().." scale = "..fMover:GetScale()..":"..f:GetScale());
  else
    RestoreFunctions(f);
  end

  if (f ~= fMover) then
    if ((isCombat and v.AH) or (not isCombat and v.HOOC)) then
      if (v.A) then f:SetAlpha(0.0) end;
      --f:EnableMouse(false);
    else
      if (v.A) then f:SetAlpha(v.A) end;
      --f:EnableMouse(true);
    end
  end
end

local function UpdateMoverPos()
  local f = _G[curFrame];
  local v = O.Frames[curFrame];
  if (f and v) then
    local n = f:GetWidth() * v.S;
    if (n < 40) then n = 40; end
    if (n > 240) then n = 240; end
    fMover:SetWidth(n);
    n = f:GetHeight() * v.S;
    if (n < 40) then n = 40; end
    if (n > 400) then n = 400; end
    fMover:SetHeight(n);
    fMover.text:SetText(v.N.."\n\nMove me\n\nRight click\nto close");
    
    UpdateFramePos(f, v);
    UpdateFramePos(fMover, v);
    
    SmartFrameMoverFrame_cbAutoHide:SetChecked(v.AH);
    SmartFrameMoverFrame_cbHideOOC:SetChecked(v.HOOC);
    SmartFrameMoverFrame_cbMoveIt:SetChecked(v.MI);
    
    if (v.MI) then
      fMover:Show();
    else
      fMover:Hide();
    end    
  end
end

local function UpdateAllFramePos(f)
  for k, v in pairs(O.Frames) do
    if (k and v) then
      if (not f or f == k) then
        UpdateFramePos(_G[k], v);
      end
    end
  end  
end


local function AddFrame(f, fn)
  if (f == nil or _G[f] == nil) then return; end  
  if (fn == nil) then fn = f; end
  -- Point, X pos, Y pos, hide in combat, move it
  
  local v = O.Frames[f];
  if (v == nil) then
    O.Frames[f] = { };
    v = O.Frames[f];
    v.N = fn;
    v.P = "TOPLEFT"; --_G[f]:GetPoint();
    v.X = 300;
    v.Y = -200;
    v.AH = false;
    v.MI = false;
    v.HOOC = false;
  end
  v.dA = _G[f]:GetAlpha();
  v.dS = _G[f]:GetScale();
  if (v.A == nil) then v.A = v.dA end
  if (v.S == nil) then v.S = v.dS end
    
  AddMsgD("Frame added: "..fn);
end


local function UpdateTrackerFrameHeight()  
	local yt, yb = UIParent:GetBottom(), UIParent:GetTop();
	--[[
	for i = 1, 50, 1 do
		local l = _G["WatchFrameLine"..i];
		if (l and l:IsVisible()) then
		  --print("Quest line = "..i);
			local t, b = l:GetTop(), l:GetBottom();
			if (t and t > yt) then yt = t; end
			if (b and b < yb) then yb = b; end
		end
	end
	local h = yt - yb + 40;
	]]--
	local h = DEFAULT_OBJECTIVE_TRACKER_MODULE[contentsHeight];
	
	if (h <= 0) then
	  h = 1;
  end
  ObjectiveTrackerFrame:SetHeight(h);
  print("ObjectiveTrackerFrame height = "..h);
end

local function AchievementAlertFrame_GetAlertFrame_ORIG()
	local name, frame, previousFrame;
	for i=1, MAX_ACHIEVEMENT_ALERTS do
		name = "AchievementAlertFrame"..i;
		frame = _G[name];
		if ( frame ) then
			if ( not frame:IsShown() ) then
				return frame;
			end
		else
			frame = CreateFrame("Button", name, UIParent, "AchievementAlertFrameTemplate");
			if ( not previousFrame ) then
				frame:SetPoint("BOTTOM", UIParent, "BOTTOM", 0, 128);
			else
				frame:SetPoint("BOTTOM", previousFrame, "TOP", 0, -10);
			end
			return frame;
		end
		previousFrame = frame;
	end
	return nil;
end

-----------------------------------------------------------------------------------------------------------------
-- END Local helper functions -----------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------

-- Init
local function Init(self)
  -- Init AchievementAlert frames
  AchievementFrame_LoadUI();
  AchievementAlertFrame_GetAlertFrame();
  
  
  fFocus = CreateFrame("Frame", nil, UIParent);
  fFocus:SetFrameStrata("DIALOG");
  fFocus:EnableMouse(false);
  fFocus:EnableKeyboard(false);
  fFocus:EnableMouseWheel(false);
  fFocus:SetSize(4, 4);

  fFocus.texture = fFocus:CreateTexture(nil, "OVERLAY");
  fFocus.texture:SetTexture(0.0, 1.0, 1.0, 0.3);
  fFocus.texture:SetAllPoints(fFocus);
  fFocus.texture:SetBlendMode("BLEND");
  
  fFocus.text = fFocus:CreateFontString(nil, nil, "Tooltip_Small");
  fFocus.text:SetJustifyH("LEFT");
  fFocus.text:SetJustifyV("TOP");
  fFocus.text:SetAllPoints(fFocus);
  
  local show = false;
  if (SmartFrameMover_Options == nil) then 
    SmartFrameMover_Options = { };
    show = true;
  end
  O = SmartFrameMover_Options;
  if (O.Debug == nil) then O.Debug = false; end  
  
  if (O.Frames == nil) then O.Frames = { }; end
  for _, v in pairs(FrameList) do
    if (v) then
      AddFrame(v[1], v[2]);
    end
  end
  
  UpdateAllFramePos();
  UpdateMoverPos();
  
  if (show) then  
    SmartFrameMover_ShowAll();
  else
    SmartFrameMover_CloseAll();
  end
  AddMsg(AddonTitleVersion.." loaded");
  AddMsg("/sfm to display movable setup frame");
  isInit = true;
end
-- END Init


-- OnLoad events
function SmartFrameMover_OnLoad(self)
  self:RegisterEvent("ADDON_LOADED");
  self:RegisterEvent("PLAYER_REGEN_DISABLED");
  self:RegisterEvent("PLAYER_REGEN_ENABLED");
	
	SlashCmdList["SmartFrameMover"] = SmartFrameMover_command;
  SLASH_SmartFrameMover1 = "/sfm";
  SLASH_SmartFrameMover2 = "/smartframemover";
  
  SlashCmdList["SmartReloadUI"] = function(msg) ReloadUI(); end;
  SLASH_SmartReloadUI1 = "/rui";
end

function SmartFrameMover_FrameOnLoad(self)
  fMain = self;
  
  self:SetMovable(true);
  self:EnableMouse(true);
  self:EnableKeyboard(false);
  self:EnableMouseWheel(false);
  self:SetClampedToScreen(false);
  --self:SetResizable(true);
  self:RegisterForDrag("LeftButton");
  
  self.texture = self:CreateTexture(nil, "BACKGROUND");
  self.texture:SetTexture(0.0, 0.0, 0.0, 0.7);      
  self.texture:SetPoint("TOPLEFT", self , "TOPLEFT", 1, -1);
  self.texture:SetPoint("TOPRIGHT", self , "TOPRIGHT", -1, -1);
  self.texture:SetPoint("BOTTOMLEFT", self , "BOTTOMLEFT", 1, 1);
  self.texture:SetPoint("BOTTOMRIGHT", self , "BOTTOMRIGHT", -2, 1);        
  self.texture:SetBlendMode("BLEND");
  self.text = self:CreateFontString(nil, nil, "GameFontHighlight");
  self.text:SetJustifyH("CENTER");
  self.text:SetJustifyV("TOP");
  self.text:SetAllPoints(self);
  self.text:SetPoint("TOPLEFT", self ,"TOPLEFT", 0, -2);
  self.text:SetPoint("BOTTOMRIGHT", self ,"BOTTOMRIGHT", 0, 0);
  self.text:SetText(AddonTitleVersion);
  
  self.info = self:CreateFontString(nil, nil, "GameFontHighlight");
  self.info:SetJustifyH("CENTER");
  self.info:SetJustifyV("BOTTOM");
  self.info:SetAllPoints(self);
  self.info:SetPoint("TOPLEFT", self ,"TOPLEFT", 0, 0);
  self.info:SetPoint("BOTTOMRIGHT", self ,"BOTTOMRIGHT", 0, 2);
  self.info:SetText("");
  
  self:ClearAllPoints();
  self:SetPoint("CENTER", UIParent, "CENTER");
end

function SmartFrameMover_MoverOnLoad(self)
  fMover = self;
  
  self:SetMovable(true);
  self:EnableMouse(true);
  self:EnableKeyboard(false);
  self:EnableMouseWheel(false);
  self:SetClampedToScreen(false);
  self:RegisterForDrag("LeftButton");
  
  self.texture = self:CreateTexture(nil, "BACKGROUND");
  self.texture:SetTexture(0.1, 0.3, 0.8, 0.6);      
  self.texture:SetPoint("TOPLEFT", self , "TOPLEFT", 1, -1);
  self.texture:SetPoint("TOPRIGHT", self , "TOPRIGHT", -1, -1);
  self.texture:SetPoint("BOTTOMLEFT", self , "BOTTOMLEFT", 1, 1);
  self.texture:SetPoint("BOTTOMRIGHT", self , "BOTTOMRIGHT", -2, 1);        
  self.texture:SetBlendMode("BLEND");
  self.text = self:CreateFontString(nil, nil, "SystemFont_Tiny"); --SystemFont_Tiny GameFontBlackTiny
  self.text:SetJustifyH("CENTER");
  self.text:SetJustifyV("MIDDLE");
  self.text:SetAllPoints(self);
  self.text:SetText("");
end
-- END OnLoad events


-- SmartFrameMover_OnEvent
local isLoaded = false;
function SmartFrameMover_OnEvent(self, event, ...)
  local arg1 = ...;
  
  --print("Event = "..tostring(event));  
  if(event == "ADDON_LOADED" and arg1 == "SmartFrameMover" and not isInit) then
    self:UnregisterEvent("ADDON_LOADED");
    isLoaded = true;
  end
  
  if (event == EventOnUpdate and isLoaded and not isInit) then
    Init(self);
  end
  
  if (not isInit) then return; end
  
  if (event == "PLAYER_REGEN_DISABLED") then
    isCombat = true;
    UpdateAllFramePos();
  end

  if (event == "PLAYER_REGEN_ENABLED") then
    isCombat = false;
    UpdateAllFramePos();
  end
  
end
-- END SmartFrameMover_OnEvent


-- SmartFrameMover_OnUpdate
local stime = 0;
local isTTreeLoaded = false;
function SmartFrameMover_OnUpdate(self, elapsed)
  if (not isInit) then    
    stime = stime + elapsed;
    --print("Update ("..stime.."sec)");
    if (stime > 0.5) then
      if (not isTTreeLoaded) then
        local _, tName = GetTalentInfo(1, 1, 1);
        if (tName) then
          --print(string.format("Talent tree ready (%.2fsec) -> Init SFM", stime));
          isTTreeLoaded = true;        
        end
      end
      if (isTTreeLoaded) then
        SmartFrameMover_OnEvent(self, EventOnUpdate);
      end      
      stime = 0;
    end
  else
    stime = stime + elapsed;
    if (fMain:IsVisible() and stime > 0.1) then
      stime = 0;
      SmartFrameMover_UpdateMouseOver();
    end
  end
end
-- END SmartFrameMover_OnUpdate


local fNameLast;
function SmartFrameMover_UpdateMouseOver()
  local fID = GetMouseFocus();
  if (not fID) then
    if (fNameLast ~= "") then
      fNameLast = "";
      fMain.info:SetText("");
    end
    return
  end
  
  local fName = fID:GetName();
  if (fName == nil or fName == fNameLast or fName == "WorldFrame" or string.find(fName, "SmartFrameMover") or string.find(fName, "DropDownList")) then return; end

  fNameLast = fName;
  local f = _G[fName];
  if (f == nil) then return; end

  fMain.info:SetText(fName);

  fFocus:ClearAllPoints();
  fFocus:SetAllPoints(f);
  fFocus.text:SetText(fName);
  fFocus:Show();
end


-- SmartFrameMover_MoverOnUpdate
function SmartFrameMover_MoverOnUpdate(self, elapsed)
  if (self.IsMoving) then
    SmartFrameMover_UpdatePos("TOPLEFT", self:GetLeft(), self:GetTop() - UIParent:GetHeight());
  end
end
-- END SmartFrameMover_MoverOnUpdate


-- SmartFrameMover_UpdatePos
function SmartFrameMover_UpdatePos(p, x, y)
  x = Round(x);
  y = Round(y);
  local f = _G[curFrame];
  local v = O.Frames[curFrame];
  if (f and v) then
    --print(curFrame..": "..p..", "..x..", "..y);
    if (v.MI) then
      --v.P = p;
      v.X = x;
      v.Y = y;
    end
    UpdateFramePos(f, v);
  end
end

function SmartFrameMover_UpdatePosExact(ox, oy)
  local v = O.Frames[curFrame];
  if (not v) then return end
  
  local x, y = Round(fMover:GetLeft() + ox), Round(fMover:GetTop() - UIParent:GetHeight() + oy);
  if (ox == 0 and oy == 0) then
    x, y = 300, -200;
    v.S = v.dS;
    v.A = v.dA;
    SmartFrameMover_UpdateSldAlpha(nil, true);
    SmartFrameMover_UpdateSldScale();
  end
  fMover:SFMClearAllPoints();
  fMover:SFMSetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y);
  --AddMsgD("UpdatePosExact: "..x..", "..y);
  SmartFrameMover_UpdatePos("TOPLEFT", x, y);
end
-- END SmartFrameMover_UpdatePos


function SmartFrameMover_GetCheckbox(cb)
  local v = O.Frames[curFrame];
  if (cb == "AH") then return v.AH; end
  if (cb == "HOOC") then return v.HOOC; end
  if (cb == "MI") then return v.MI; end
  return false;
end

function SmartFrameMover_SetCheckbox(cb, b)
  local v = O.Frames[curFrame];
  if (cb == "AH") then v.AH = b; return; end
  if (cb == "HOOC") then v.HOOC = b; UpdateFramePos(_G[curFrame], v); return; end
  if (cb == "MI") then v.MI = b; UpdateMoverPos(); return; end
end

function SmartFrameMover_SliderOnLoad(self, low, high, step)
  self:SetScale(0.92);
  _G[self:GetName().."Low"]:SetText("");
  _G[self:GetName().."High"]:SetText("");
  self:SetMinMaxValues(low, high);
  self:SetValueStep(step);
end

function SmartFrameMover_UpdateSldScale(v, noUd)
  if (not O) then return end
  local fv = O.Frames[curFrame];
  local f = fMain.scale;
  
  if (v) then
    if (v == fv.S) then return end
    fv.S = v;
  else
    v = fv.S;
    f:SetValue(v);
  end
  
  _G[f:GetName().."Text"]:SetFormattedText("Scale: %s%.2f%s", GFN, v, "|r");
  if (not noUd) then
    UpdateFramePos(_G[curFrame], fv);
    UpdateMoverPos();
  end
end

function SmartFrameMover_UpdateSldAlpha(v, noUd)
  if (not O) then return end
  local fv = O.Frames[curFrame];
  local f = fMain.alpha;
  
  if (v) then
    if (v == fv.A) then return end
    fv.A = v;
  else
    v = fv.A;
    f:SetValue(v);
  end
  
  _G[f:GetName().."Text"]:SetFormattedText("Alpha: %s%.2f%s", GFN, v, "|r");
  if (not noUd) then UpdateFramePos(_G[curFrame], fv) end
end


function SmartFrameMover_OnShow(self)
  self:RegisterEvent("PLAYER_REGEN_ENABLED");
end

function SmartFrameMover_ShowAll()
  UpdateAllFramePos();
  UpdateMoverPos();
  fMain:Show();
  fMover:Show();
  if (O.Debug) then
    Boss1TargetFrame:Show();
  end
end

function SmartFrameMover_CloseAll()
  fMain:Hide();
  fMover:Hide();
  fFocus:Hide();
  if (O.Debug) then
    Boss1TargetFrame:Hide();
  end
end


-- Dropdown methods
function SmartFrameMover_DropDownOnShow(self)
  local i = 0;
  for _, v in ipairs(FrameList) do
    i = i + 1;
    --AddMsgD(i .. "." .. k);
    if (v and v[1] == curFrame) then
      break;
    end
  end
  UIDropDownMenu_Initialize(self, SmartFrameMover_DropDownInitialize);
  UIDropDownMenu_SetSelectedValue(SmartFrameMoverFrame_ddFrame, i);
  UIDropDownMenu_SetWidth(SmartFrameMoverFrame_ddFrame, 135);
end

function SmartFrameMover_DropDownInitialize()
  local info = UIDropDownMenu_CreateInfo();
	info.text = ALL;
	info.value = -1;
	info.func = SmartFrameMover_DropDownOnClick;
  for i, v in ipairs(FrameList) do
    if (v and (v[1] == "-" or _G[v[1]])) then
      info.value = i;
      if (v[1] == "-") then
        info.text = "------------------------------";
        --info.disabled = 1;
        info.notClickable = 1;
        info.notCheckable = 1;
      else
        info.text = v[2];
        info.func = SmartFrameMover_DropDownOnClick;
        info.disabled = nil;
        info.notClickable = nil;
        info.notCheckable = nil;
      end
      info.checked = nil;
      UIDropDownMenu_AddButton(info);
    end
  end
end

function SmartFrameMover_DropDownOnClick(self)
  local n = self.value;  
  --AddMsgD("DD value = "..i);
  UIDropDownMenu_SetSelectedValue(SmartFrameMoverFrame_ddFrame, n);
  local t;
  for i, v in ipairs(FrameList) do
    if (i == n) then
      t = v[1];
      break;
    end
  end
  if (t and curFrame ~= t) then
    --AddMsgD("Frame = "..t);
    curFrame = t;
    SmartFrameMover_UpdateSldAlpha(nil, true);
    SmartFrameMover_UpdateSldScale();
    UpdateMoverPos();
  end
end
-- END Dropdown methods


-----------------------------------------------------------------------------------------------------------------
-- Command line menu --------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------

function SmartFrameMover_command(msgIn)
  if (not isInit) then return end
  
  local msgs = Split(msgIn, " ");
  local msg = msgs[1];
  
  if (msg == nil or msg == "") then
    if(fMain:IsVisible()) then
      SmartFrameMover_CloseAll();
    else
      SmartFrameMover_ShowAll();
    end
  elseif (msg == "hv") then
    --O.HideV = BToggle(O.HideV, "Hide vehicle frame always = ");
  elseif (msg == "debug") then
    O.Debug = BToggle(O.Debug, "Debug = ");
  elseif (msg == "reset") then
    SmartFrameMover_CloseAll();
    SmartFrameMover_Options = nil;
    Init(fMain);
    AddMsg(AddonTitle..": Set all to default");
  elseif (msg == "?" or msg == "help") then
    AddMsg(AddonTitleVersion, true);
    AddMsg("Syntax: /sfm [command] or /smartframemover [command]");
    --AddMsg("hv    -  " .. "Hide vehicle frame always = "..BToStr(O.HideV));
    AddMsg("reset -  " .. "Reset all to default");  
  end
end
-- END SmartFrameMover_command


-----------------------------------------------------------------------------------------------------------------
-- Hooks --------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------

-- Auto resize for the quest watch frame
local isLockWF = false;
function SmartFrameMover_ObjectiveTracker_Update(...)
  if (isLockWF) then return; end
  
  isLockWF = true;
  local h = UIParent:GetTop();
  h = h - (h - ObjectiveTrackerFrame:GetTop());
  
  --print(h);
  --print(ObjectiveTrackerFrame.BlocksFrame.contentsHeight);
  
  ObjectiveTrackerFrame:SetHeight(h);
  
  HookObjectiveTracker_Update(...);
  --UpdateTrackerFrameHeight();
  --HookObjectiveTracker_Update(...);
  isLockWF = false;
end

ObjectiveTracker_Update = SmartFrameMover_ObjectiveTracker_Update;
